# Copyright 2019-2025 SURF
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Simple search migration function that runs when SEARCH_ENABLED = True."""

import pydantic_ai
from sqlalchemy import text
from sqlalchemy.engine import Connection
from structlog import get_logger

from orchestrator.llm_settings import llm_settings
from orchestrator.search.core.types import FieldType

logger = get_logger(__name__)

TABLE = "ai_search_index"
TARGET_DIM = 1536


def run_migration(connection: Connection) -> None:
    """Run LLM migration with ON CONFLICT DO NOTHING pattern."""
    logger.info("Running LLM migration", pydantic_ai_version=pydantic_ai.__version__)

    try:
        # Test to see if the extenstion exists and then skip the migration; Needed for certain situations where db user
        # has insufficient priviledges to run the `CREATE EXTENSION ...` command.
        res = connection.execute(text("SELECT * FROM pg_extension where extname = 'vector';"))
        if llm_settings.LLM_FORCE_EXTENTION_MIGRATION or res.rowcount == 0:
            # Create PostgreSQL extensions
            logger.info("Attempting to run the extention creation;")
            connection.execute(text('CREATE EXTENSION IF NOT EXISTS "uuid-ossp";'))
            connection.execute(text("CREATE EXTENSION IF NOT EXISTS ltree;"))
            connection.execute(text("CREATE EXTENSION IF NOT EXISTS unaccent;"))
            connection.execute(text("CREATE EXTENSION IF NOT EXISTS pg_trgm;"))
            connection.execute(text("CREATE EXTENSION IF NOT EXISTS vector;"))

        # Create field_type enum
        field_type_values = "', '".join([ft.value for ft in FieldType])
        connection.execute(
            text(
                f"""
            DO $$
            BEGIN
                IF NOT EXISTS (SELECT 1 FROM pg_type WHERE typname = 'field_type') THEN
                    CREATE TYPE field_type AS ENUM ('{field_type_values}');
                END IF;
            END $$;
        """
            )
        )

        # Create table with ON CONFLICT DO NOTHING pattern
        connection.execute(
            text(
                f"""
            CREATE TABLE IF NOT EXISTS {TABLE} (
                entity_type TEXT NOT NULL,
                entity_id UUID NOT NULL,
                entity_title TEXT,
                path LTREE NOT NULL,
                value TEXT NOT NULL,
                embedding VECTOR({TARGET_DIM}),
                content_hash VARCHAR(64) NOT NULL,
                value_type field_type NOT NULL DEFAULT '{FieldType.STRING.value}',
                CONSTRAINT pk_ai_search_index PRIMARY KEY (entity_id, path)
            );
        """
            )
        )

        # Drop default
        connection.execute(text(f"ALTER TABLE {TABLE} ALTER COLUMN value_type DROP DEFAULT;"))

        # Add entity_title column if it doesn't exist (for existing installations)
        connection.execute(
            text(
                f"""
            DO $$
            BEGIN
                IF NOT EXISTS (
                    SELECT 1 FROM information_schema.columns
                    WHERE table_name = '{TABLE}' AND column_name = 'entity_title'
                ) THEN
                    ALTER TABLE {TABLE} ADD COLUMN entity_title TEXT;
                END IF;
            END $$;
        """
            )
        )

        # Create indexes with IF NOT EXISTS
        connection.execute(text(f"CREATE INDEX IF NOT EXISTS ix_ai_search_index_entity_id ON {TABLE} (entity_id);"))
        connection.execute(
            text(f"CREATE INDEX IF NOT EXISTS idx_ai_search_index_content_hash ON {TABLE} (content_hash);")
        )
        connection.execute(
            text(f"CREATE INDEX IF NOT EXISTS ix_flat_path_gist ON {TABLE} USING GIST (path gist_ltree_ops);")
        )
        connection.execute(text(f"CREATE INDEX IF NOT EXISTS ix_flat_path_btree ON {TABLE} (path);"))
        connection.execute(
            text(f"CREATE INDEX IF NOT EXISTS ix_flat_value_trgm ON {TABLE} USING GIN (value gin_trgm_ops);")
        )
        connection.execute(
            text(
                f"CREATE INDEX IF NOT EXISTS ix_flat_embed_hnsw ON {TABLE} USING HNSW (embedding vector_l2_ops) WITH (m = 16, ef_construction = 64);"
            )
        )

        # Create agent_runs table
        connection.execute(
            text(
                """
            CREATE TABLE IF NOT EXISTS agent_runs (
                run_id UUID PRIMARY KEY DEFAULT uuid_generate_v4(),
                agent_type VARCHAR(50) NOT NULL,
                created_at TIMESTAMP WITH TIME ZONE DEFAULT CURRENT_TIMESTAMP NOT NULL
            );
        """
            )
        )
        connection.execute(text("CREATE INDEX IF NOT EXISTS ix_agent_runs_created_at ON agent_runs (created_at);"))

        # Create search_queries table
        connection.execute(
            text(
                f"""
            CREATE TABLE IF NOT EXISTS search_queries (
                query_id UUID PRIMARY KEY DEFAULT uuid_generate_v4(),
                run_id UUID,
                query_number INTEGER NOT NULL,
                parameters JSONB NOT NULL,
                query_embedding VECTOR({TARGET_DIM}),
                executed_at TIMESTAMP WITH TIME ZONE DEFAULT CURRENT_TIMESTAMP NOT NULL,
                CONSTRAINT fk_search_queries_run_id FOREIGN KEY (run_id) REFERENCES agent_runs(run_id) ON DELETE CASCADE
            );
        """
            )
        )
        connection.execute(text("CREATE INDEX IF NOT EXISTS ix_search_queries_run_id ON search_queries (run_id);"))
        connection.execute(
            text("CREATE INDEX IF NOT EXISTS ix_search_queries_executed_at ON search_queries (executed_at);")
        )
        connection.execute(text("CREATE INDEX IF NOT EXISTS ix_search_queries_query_id ON search_queries (query_id);"))

        connection.commit()
        logger.info("LLM migration completed successfully")

    except Exception as e:
        logger.error("LLM migration failed", error=str(e))
        raise Exception(
            f"LLM migration failed. This likely means the pgvector extension "
            f"is not installed. Please install pgvector and ensure your PostgreSQL "
            f"version supports it. Error: {e}"
        ) from e



# Copyright 2019-2025 SURF, GÉANT.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
from typing import Any, cast

import structlog
from ag_ui.core import EventType, StateSnapshotEvent
from pydantic import ValidationError
from pydantic_ai import RunContext
from pydantic_ai.ag_ui import StateDeps
from pydantic_ai.exceptions import ModelRetry
from pydantic_ai.messages import ModelRequest, UserPromptPart
from pydantic_ai.toolsets import FunctionToolset

from orchestrator.api.api_v1.endpoints.search import (
    get_definitions,
    list_paths,
)
from orchestrator.db import db
from orchestrator.search.agent.handlers import (
    execute_aggregation_with_persistence,
    execute_search_with_persistence,
)
from orchestrator.search.agent.state import SearchState
from orchestrator.search.agent.validation import require_action
from orchestrator.search.aggregations import Aggregation, FieldAggregation, TemporalGrouping
from orchestrator.search.core.types import ActionType, EntityType, FilterOp
from orchestrator.search.filters import FilterTree
from orchestrator.search.query import engine
from orchestrator.search.query.exceptions import PathNotFoundError, QueryValidationError
from orchestrator.search.query.export import fetch_export_data
from orchestrator.search.query.mixins import OrderBy
from orchestrator.search.query.queries import AggregateQuery, CountQuery, Query, SelectQuery
from orchestrator.search.query.results import AggregationResponse, AggregationResult, ExportData, VisualizationType
from orchestrator.search.query.state import QueryState
from orchestrator.search.query.validation import (
    validate_aggregation_field,
    validate_filter_tree,
    validate_grouping_fields,
    validate_order_by_fields,
    validate_temporal_grouping_field,
)
from orchestrator.settings import app_settings

logger = structlog.get_logger(__name__)

search_toolset: FunctionToolset[StateDeps[SearchState]] = FunctionToolset(max_retries=1)


def last_user_message(ctx: RunContext[StateDeps[SearchState]]) -> str | None:
    for msg in reversed(ctx.messages):
        if isinstance(msg, ModelRequest):
            for part in msg.parts:
                if isinstance(part, UserPromptPart) and isinstance(part.content, str):
                    return part.content
    return None


@search_toolset.tool
async def start_new_search(
    ctx: RunContext[StateDeps[SearchState]],
    entity_type: EntityType,
    action: ActionType = ActionType.SELECT,
) -> StateSnapshotEvent:
    """Starts a completely new search, clearing all previous state.

    This MUST be the first tool called when the user asks for a NEW search.
    Warning: This will erase any existing filters, results, and search state.
    """
    final_query = last_user_message(ctx) or ""

    logger.debug(
        "Starting new search",
        entity_type=entity_type.value,
        action=action,
        query=final_query,
    )

    # Clear all state
    ctx.deps.state.results_count = None
    ctx.deps.state.action = action

    # Create the appropriate query object based on action
    if action == ActionType.SELECT:
        ctx.deps.state.query = SelectQuery(
            entity_type=entity_type,
            query_text=final_query,
        )
    elif action == ActionType.COUNT:
        ctx.deps.state.query = CountQuery(
            entity_type=entity_type,
        )
    else:  # ActionType.AGGREGATE
        ctx.deps.state.query = AggregateQuery(
            entity_type=entity_type,
            aggregations=[],  # Will be set by set_aggregations tool
        )

    logger.debug("New search started", action=action.value, query_type=type(ctx.deps.state.query).__name__)

    return StateSnapshotEvent(
        type=EventType.STATE_SNAPSHOT,
        snapshot=ctx.deps.state.model_dump(),
    )


@search_toolset.tool(retries=2)
async def set_filter_tree(
    ctx: RunContext[StateDeps[SearchState]],
    filters: FilterTree | None,
) -> StateSnapshotEvent:
    """Replace current filters atomically with a full FilterTree, or clear with None.

    See FilterTree model for structure, operators, and examples.
    """
    if ctx.deps.state.query is None:
        raise ModelRetry("Search query is not initialized. Call start_new_search first.")

    entity_type = ctx.deps.state.query.entity_type

    logger.debug(
        "Setting filter tree",
        entity_type=entity_type.value,
        has_filters=filters is not None,
        filter_summary=f"{len(filters.get_all_leaves())} filters" if filters else "no filters",
    )

    try:
        await validate_filter_tree(filters, entity_type)
    except PathNotFoundError as e:
        logger.debug(f"{PathNotFoundError.__name__}: {str(e)}")
        raise ModelRetry(f"{str(e)} Use discover_filter_paths tool to find valid paths.")
    except QueryValidationError as e:
        # ModelRetry will trigger an agent retry, containing the specific validation error.
        logger.debug(f"Query validation failed: {str(e)}")
        raise ModelRetry(str(e))
    except Exception as e:
        logger.error("Unexpected Filter validation exception", error=str(e))
        raise ModelRetry(f"Filter validation failed: {str(e)}. Please check your filter structure and try again.")

    ctx.deps.state.query = cast(Query, ctx.deps.state.query).model_copy(update={"filters": filters})

    # Use snapshot to workaround state persistence issue
    # TODO: Fix root cause; state tree may be empty on frontend when parameters are being set
    return StateSnapshotEvent(
        type=EventType.STATE_SNAPSHOT,
        snapshot=ctx.deps.state.model_dump(),
    )


@search_toolset.tool
@require_action(ActionType.SELECT)
async def run_search(
    ctx: RunContext[StateDeps[SearchState]],
    limit: int = 10,
) -> AggregationResponse:
    """Execute a search to find and rank entities.

    Use this tool for SELECT action to find entities matching your criteria.
    For counting or computing statistics, use run_aggregation instead.
    """
    query = cast(SelectQuery, cast(Query, ctx.deps.state.query).model_copy(update={"limit": limit}))

    search_response, run_id, query_id = await execute_search_with_persistence(query, db.session, ctx.deps.state.run_id)

    ctx.deps.state.run_id = run_id
    ctx.deps.state.query_id = query_id
    ctx.deps.state.results_count = len(search_response.results)

    # Convert SearchResults to AggregationResults for consistent rendering
    aggregation_results = [
        AggregationResult(
            group_values={
                "entity_id": result.entity_id,
                "title": result.entity_title,
                "entity_type": result.entity_type.value,
            },
            aggregations={"score": result.score},
        )
        for result in search_response.results
    ]

    # For now use the default table visualization for search results
    aggregation_response = AggregationResponse(
        results=aggregation_results,
        total_groups=len(aggregation_results),
        metadata=search_response.metadata,
        visualization_type=VisualizationType(type="table"),
    )

    logger.debug(
        "Search completed",
        total_count=ctx.deps.state.results_count,
        query_id=str(query_id),
    )

    return aggregation_response


@search_toolset.tool
@require_action(ActionType.COUNT, ActionType.AGGREGATE)
async def run_aggregation(
    ctx: RunContext[StateDeps[SearchState]],
    visualization_type: VisualizationType,
) -> AggregationResponse:
    """Execute an aggregation to compute counts or statistics over entities.

    Use this tool for COUNT or AGGREGATE actions after setting up:
    - Grouping fields with set_grouping or set_temporal_grouping
    - Aggregation functions with set_aggregations (for AGGREGATE action)
    """
    query = cast(CountQuery | AggregateQuery, ctx.deps.state.query)

    logger.debug(
        "Executing aggregation",
        search_entity_type=query.entity_type.value,
        has_filters=query.filters is not None,
        action=query.action,
    )

    aggregation_response, run_id, query_id = await execute_aggregation_with_persistence(
        query, db.session, ctx.deps.state.run_id
    )

    ctx.deps.state.run_id = run_id
    ctx.deps.state.query_id = query_id
    ctx.deps.state.results_count = aggregation_response.total_groups

    aggregation_response.visualization_type = visualization_type

    logger.debug(
        "Aggregation completed",
        total_groups=aggregation_response.total_groups,
        visualization_type=visualization_type.type,
        query_id=str(query_id),
    )

    return aggregation_response


@search_toolset.tool
async def discover_filter_paths(
    ctx: RunContext[StateDeps[SearchState]],
    field_names: list[str],
    entity_type: EntityType | None = None,
) -> dict[str, dict[str, Any]]:
    """Discovers available filter paths for a list of field names.

    Returns a dictionary where each key is a field_name from the input list and
    the value is its discovery result.
    """
    if not entity_type:
        if ctx.deps.state.query:
            entity_type = ctx.deps.state.query.entity_type
        else:
            raise ModelRetry("Entity type not specified and no query in state. Call start_new_search first.")

    all_results = {}
    for field_name in field_names:
        paths_response = await list_paths(prefix="", q=field_name, entity_type=entity_type, limit=100)

        matching_leaves = []
        for leaf in paths_response.leaves:
            if field_name.lower() in leaf.name.lower():
                matching_leaves.append(
                    {
                        "name": leaf.name,
                        "value_kind": leaf.ui_types,
                        "paths": leaf.paths,
                    }
                )

        matching_components = []
        for comp in paths_response.components:
            if field_name.lower() in comp.name.lower():
                matching_components.append(
                    {
                        "name": comp.name,
                        "value_kind": comp.ui_types,
                    }
                )

        result_for_field: dict[str, Any]
        if not matching_leaves and not matching_components:
            result_for_field = {
                "status": "NOT_FOUND",
                "guidance": f"No filterable paths found containing '{field_name}'. Do not create a filter for this.",
                "leaves": [],
                "components": [],
            }
        else:
            result_for_field = {
                "status": "OK",
                "guidance": f"Found {len(matching_leaves)} field(s) and {len(matching_components)} component(s) for '{field_name}'.",
                "leaves": matching_leaves,
                "components": matching_components,
            }

        all_results[field_name] = result_for_field
    logger.debug("Returning found fieldname - path mapping", all_results=all_results)
    return all_results


@search_toolset.tool
async def get_valid_operators() -> dict[str, list[FilterOp]]:
    """Gets the mapping of field types to their valid filter operators."""
    definitions = await get_definitions()

    operator_map = {}
    for ui_type, type_def in definitions.items():
        key = ui_type.value

        if hasattr(type_def, "operators"):
            operator_map[key] = type_def.operators
    return operator_map


@search_toolset.tool
async def fetch_entity_details(
    ctx: RunContext[StateDeps[SearchState]],
    limit: int = 10,
) -> str:
    """Fetch detailed entity information to answer user questions.

    Use this tool when you need detailed information about entities from the search results
    to answer the user's question. This provides the same detailed data that would be
    included in an export (e.g., subscription status, product details, workflow info, etc.).

    Args:
        ctx: Runtime context for agent (injected).
        limit: Maximum number of entities to fetch details for (default 10).

    Returns:
        JSON string containing detailed entity information.

    Raises:
        ModelRetry: If no search has been executed.
    """
    if ctx.deps.state.query_id is None:
        raise ModelRetry("No query_id found. Run a search first.")

    # Load the saved query and re-execute it to get entity IDs
    query_state = QueryState.load_from_id(ctx.deps.state.query_id, SelectQuery)
    query = query_state.query.model_copy(update={"limit": limit})
    search_response = await engine.execute_search(query, db.session)
    entity_ids = [r.entity_id for r in search_response.results]

    if not entity_ids:
        return json.dumps({"message": "No entities found in search results."})

    entity_type = query.entity_type

    logger.debug(
        "Fetching detailed entity data",
        entity_type=entity_type.value,
        entity_count=len(entity_ids),
    )

    detailed_data = fetch_export_data(entity_type, entity_ids)

    return json.dumps(detailed_data, indent=2)


@search_toolset.tool
@require_action(ActionType.SELECT)
async def prepare_export(
    ctx: RunContext[StateDeps[SearchState]],
) -> ExportData:
    """Prepares export URL using the last executed search query.

    Returns export data which is displayed directly in the UI.
    """
    if not ctx.deps.state.query_id or not ctx.deps.state.run_id:
        raise ModelRetry("No search has been executed yet. Run a search first before exporting.")

    logger.debug(
        "Prepared query for export",
        query_id=str(ctx.deps.state.query_id),
    )

    download_url = f"{app_settings.BASE_URL}/api/search/queries/{ctx.deps.state.query_id}/export"

    export_data = ExportData(
        query_id=str(ctx.deps.state.query_id),
        download_url=download_url,
        message="Export ready for download.",
    )

    logger.debug("Export prepared", query_id=export_data.query_id)

    return export_data


@search_toolset.tool(retries=2)
@require_action(ActionType.COUNT, ActionType.AGGREGATE)
async def set_grouping(
    ctx: RunContext[StateDeps[SearchState]],
    group_by_paths: list[str],
    order_by: list[OrderBy] | None = None,
) -> StateSnapshotEvent:
    """Set which field paths to group results by for aggregation.

    Only used with COUNT or AGGREGATE actions. Paths must exist in the schema; use discover_filter_paths to verify.
    Optionally specify ordering for the grouped results.

    For order_by: You can order by grouping field paths OR aggregation aliases (e.g., 'count').
    Grouping field paths will be validated; aggregation aliases cannot be validated until execution.
    """
    try:
        validate_grouping_fields(group_by_paths)
        validate_order_by_fields(order_by)
    except PathNotFoundError as e:
        raise ModelRetry(f"{str(e)} Use discover_filter_paths to find valid paths.")

    update_dict: dict[str, Any] = {"group_by": group_by_paths}
    if order_by is not None:
        update_dict["order_by"] = order_by

    try:
        ctx.deps.state.query = cast(Query, ctx.deps.state.query).model_copy(update=update_dict)
    except ValidationError as e:
        raise ModelRetry(str(e))

    return StateSnapshotEvent(
        type=EventType.STATE_SNAPSHOT,
        snapshot=ctx.deps.state.model_dump(),
    )


@search_toolset.tool(retries=2)
@require_action(ActionType.AGGREGATE)
async def set_aggregations(
    ctx: RunContext[StateDeps[SearchState]],
    aggregations: list[Aggregation],
) -> StateSnapshotEvent:
    """Define what aggregations to compute over the matching records.

    Only used with AGGREGATE action. See Aggregation model (CountAggregation, FieldAggregation) for structure and field requirements.

    """
    # Validate field paths for FieldAggregations
    try:
        for agg in aggregations:
            if isinstance(agg, FieldAggregation):
                validate_aggregation_field(agg.type, agg.field)
    except PathNotFoundError as e:
        raise ModelRetry(
            f"{str(e)} "
            f"You MUST call discover_filter_paths first to find valid fields. "
            f"If the field truly doesn't exist, inform the user that this data is not available."
        )
    except QueryValidationError as e:
        raise ModelRetry(f"{str(e)}")

    try:
        ctx.deps.state.query = cast(Query, ctx.deps.state.query).model_copy(update={"aggregations": aggregations})
    except ValidationError as e:
        raise ModelRetry(str(e))

    return StateSnapshotEvent(
        type=EventType.STATE_SNAPSHOT,
        snapshot=ctx.deps.state.model_dump(),
    )


@search_toolset.tool(retries=2)
@require_action(ActionType.COUNT, ActionType.AGGREGATE)
async def set_temporal_grouping(
    ctx: RunContext[StateDeps[SearchState]],
    temporal_groups: list[TemporalGrouping],
    cumulative: bool = False,
    order_by: list[OrderBy] | None = None,
) -> StateSnapshotEvent:
    """Set temporal grouping to group datetime fields by time periods.

    Only used with COUNT or AGGREGATE actions. See TemporalGrouping model for structure, periods, and examples.
    Optionally enable cumulative aggregations (running totals) and specify ordering.

    For order_by: You can order by temporal field paths OR aggregation aliases (e.g., 'count').
    Temporal field paths will be validated; aggregation aliases cannot be validated until execution.
    """
    try:
        for tg in temporal_groups:
            validate_temporal_grouping_field(tg.field)
        validate_order_by_fields(order_by)
    except PathNotFoundError as e:
        raise ModelRetry(f"{str(e)} Use discover_filter_paths to find valid paths.")
    except QueryValidationError as e:
        raise ModelRetry(f"{str(e)} Use discover_filter_paths to find datetime fields.")

    update_dict: dict[str, Any] = {"temporal_group_by": temporal_groups}
    if cumulative:
        update_dict["cumulative"] = cumulative
    if order_by is not None:
        update_dict["order_by"] = order_by

    try:
        ctx.deps.state.query = cast(Query, ctx.deps.state.query).model_copy(update=update_dict)
    except ValidationError as e:
        raise ModelRetry(str(e))

    return StateSnapshotEvent(
        type=EventType.STATE_SNAPSHOT,
        snapshot=ctx.deps.state.model_dump(),
    )



"""Create linker table workflow_apscheduler.

Revision ID: 961eddbd4c13
Revises: 850dccac3b02
Create Date: 2025-11-18 10:38:57.211087

"""

from uuid import uuid4

import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "961eddbd4c13"
down_revision = "850dccac3b02"
branch_labels = None
depends_on = None

# NOTE: this migration forgot to insert these workflows with is_task=true. Make sure to correct that if you copy this.
workflows = [
    {
        "name": "task_validate_subscriptions",
        "description": "Validate subscriptions",
        "workflow_id": uuid4(),
        "target": "SYSTEM",
    },
]


def _create_workflows() -> None:
    conn = op.get_bind()
    for workflow in workflows:
        conn.execute(
            sa.text(
                "INSERT INTO workflows VALUES (:workflow_id, :name, :target, :description, now()) ON CONFLICT DO NOTHING"
            ),
            workflow,
        )


def _downgrade_create_workflows() -> None:
    conn = op.get_bind()
    for workflow in workflows:
        conn.execute(sa.text("DELETE FROM workflows WHERE name = :name"), {"name": workflow["name"]})


def _create_apscheduler_jobs_table_if_not_exists() -> None:
    # Check if the apscheduler_jobs table exists and create it if it does not exist.
    conn = op.get_bind()
    inspector = sa.inspect(conn)
    if "apscheduler_jobs" not in inspector.get_table_names():
        op.execute(
            sa.text(
                """
                CREATE TABLE apscheduler_jobs
                (
                    id            VARCHAR(191) NOT NULL PRIMARY KEY,
                    next_run_time DOUBLE PRECISION,
                    job_state     bytea NOT NULL
                );
                """
            )
        )


def _create_workflows_table_if_not_exists() -> None:
    # Notice the VARCHAR(512) for schedule_id to accommodate longer IDs
    # This so that if APScheduler changes its ID format in the future, we are covered.
    op.execute(
        sa.text(
            """
                CREATE TABLE workflows_apscheduler_jobs (
                    workflow_id UUID NOT NULL,
                    schedule_id VARCHAR(512) NOT NULL,
                    PRIMARY KEY (workflow_id, schedule_id),
                    CONSTRAINT fk_workflow
                        FOREIGN KEY (workflow_id) REFERENCES public.workflows (workflow_id)
                            ON DELETE CASCADE,
                    CONSTRAINT fk_schedule
                        FOREIGN KEY (schedule_id) REFERENCES public.apscheduler_jobs (id)
                            ON DELETE CASCADE,
                    CONSTRAINT uq_workflow_schedule UNIQUE (workflow_id, schedule_id)
                );
            """
        )
    )

    op.create_index("ix_workflows_apscheduler_jobs_schedule_id", "workflows_apscheduler_jobs", ["schedule_id"])


def upgrade() -> None:
    _create_apscheduler_jobs_table_if_not_exists()
    _create_workflows_table_if_not_exists()
    _create_workflows()


def downgrade() -> None:
    op.execute(
        sa.text(
            """
            DROP TABLE IF EXISTS workflows_apscheduler_jobs;
            """
        )
    )
    _downgrade_create_workflows()



# Copyright 2022-2023 SURF, GÉANT.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from collections.abc import Generator
from itertools import count
from typing import TYPE_CHECKING, Annotated, Any
from uuid import UUID

import strawberry
from pydantic.alias_generators import to_camel as to_lower_camel
from strawberry.scalars import JSON

from orchestrator.graphql.schemas.product_block import owner_subscription_resolver
from orchestrator.graphql.types import OrchestratorInfo
from orchestrator.graphql.utils.get_selected_paths import get_selected_paths
from orchestrator.utils.get_subscription_dict import get_subscription_dict

if TYPE_CHECKING:
    from orchestrator.graphql.schemas.subscription import SubscriptionInterface


@strawberry.type
class ProductBlockInstance:
    id: int
    parent: int | None
    subscription_instance_id: UUID
    owner_subscription_id: UUID
    in_use_by_relations: list[JSON]
    product_block_instance_values: JSON
    subscription: (
        Annotated["SubscriptionInterface", strawberry.lazy("orchestrator.graphql.schemas.subscription")] | None
    ) = strawberry.field(
        description="resolve to subscription of the product block", resolver=owner_subscription_resolver
    )


def is_product_block(candidate: Any) -> bool:
    if isinstance(candidate, dict):
        # TODO: also filter on tag (needs addition of tag in orchestrator endpoint)
        # NOTE: this crosses subscription boundaries. If needed we can add an additional filter to limit that.
        return candidate.get("owner_subscription_id", None)
    return False


def get_all_product_blocks(subscription: dict[str, Any], _tags: list[str] | None) -> list[dict[str, Any]]:
    gen_id = count()

    def locate_product_block(candidate: dict[str, Any]) -> Generator:
        def new_product_block(item: dict[str, Any]) -> Generator:
            enriched_item = item | {"id": next(gen_id), "parent": candidate.get("id")}
            yield enriched_item
            yield from locate_product_block(enriched_item)

        for value in candidate.values():
            if is_product_block(value):
                yield from new_product_block(value)
            elif isinstance(value, list):
                for item in value:
                    if is_product_block(item):
                        yield from new_product_block(item)

    return list(locate_product_block(subscription))


pb_instance_property_keys = (
    "id",
    "parent",
    "owner_subscription_id",
    "subscription_instance_id",
    "in_use_by_relations",
    "in_use_by_ids",
)


async def get_subscription_product_blocks(
    info: OrchestratorInfo,
    subscription_id: UUID,
    tags: list[str] | None = None,
    product_block_instance_values: list[str] | None = None,
) -> list[ProductBlockInstance]:
    inject_inuseby = "in_use_by_relations" in get_selected_paths(info)
    subscription, _ = await get_subscription_dict(subscription_id, inject_inuseby=inject_inuseby)

    def to_product_block(product_block: dict[str, Any]) -> ProductBlockInstance:
        def is_resource_type(candidate: Any) -> bool:
            return not isinstance(candidate, dict)

        def requested_resource_type(key: str) -> bool:
            return not product_block_instance_values or key in product_block_instance_values

        def included(key: str, value: Any) -> bool:
            return is_resource_type(value) and requested_resource_type(key) and key not in pb_instance_property_keys

        def value_parser(value: Any) -> str | int | float | list | None:
            if isinstance(value, (str, int, float, type(None))):
                return value

            if isinstance(value, list):
                return [value_parser(v) for v in value if is_resource_type(v)]

            return str(value)

        return ProductBlockInstance(
            id=product_block["id"],
            parent=product_block.get("parent"),
            owner_subscription_id=product_block["owner_subscription_id"],
            subscription_instance_id=product_block["subscription_instance_id"],
            product_block_instance_values=[
                {"field": to_lower_camel(k), "value": value_parser(v)}
                for k, v in product_block.items()
                if included(k, v)
            ],
            in_use_by_relations=product_block.get("in_use_by_relations", []),
        )

    product_blocks = (to_product_block(product_block) for product_block in get_all_product_blocks(subscription, tags))
    return [product_block for product_block in product_blocks if product_block.product_block_instance_values]



# Copyright 2019-2025 SURF, GÉANT.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import structlog
from fastapi import APIRouter, HTTPException, Query, status

from orchestrator.db import db
from orchestrator.schemas.search import (
    CursorInfoSchema,
    ExportResponse,
    PageInfoSchema,
    PathsResponse,
    SearchResultsSchema,
)
from orchestrator.schemas.search_requests import SearchRequest
from orchestrator.search.core.exceptions import InvalidCursorError, QueryStateNotFoundError
from orchestrator.search.core.types import EntityType, UIType
from orchestrator.search.filters.definitions import TypeDefinition, generate_definitions
from orchestrator.search.query import QueryState, engine
from orchestrator.search.query.builder import build_paths_query, create_path_autocomplete_lquery, process_path_rows
from orchestrator.search.query.queries import ExportQuery, SelectQuery
from orchestrator.search.query.results import SearchResult
from orchestrator.search.query.validation import is_lquery_syntactically_valid
from orchestrator.search.retrieval.pagination import PageCursor, encode_next_page_cursor

router = APIRouter()
logger = structlog.get_logger(__name__)


async def _perform_search_and_fetch(
    entity_type: EntityType | None = None,
    request: SearchRequest | None = None,
    cursor: str | None = None,
    query_id: str | None = None,
) -> SearchResultsSchema[SearchResult]:
    """Execute search with optional pagination.

    Args:
        entity_type: Entity type to search
        request: Search request for new search
        cursor: Pagination cursor (loads saved query state)
        query_id: Saved query ID to retrieve and execute

    Returns:
        Search results with entity_id, score, and matching_field.
    """
    try:
        page_cursor: PageCursor | None = None
        query: SelectQuery

        if cursor:
            page_cursor = PageCursor.decode(cursor)
            query_state = QueryState.load_from_id(page_cursor.query_id, SelectQuery)
            query = query_state.query

        elif query_id:
            query_state = QueryState.load_from_id(query_id, SelectQuery)
            query = query_state.query

        elif request and entity_type:
            query = request.to_query(entity_type)
            query_state = QueryState(query=query, query_embedding=None)
        else:
            raise HTTPException(
                status_code=status.HTTP_400_BAD_REQUEST,
                detail="Either (request + entity_type), cursor, or query_id must be provided",
            )

        search_response = await engine.execute_search(query, db.session, page_cursor, query_state.query_embedding)
        if not search_response.results:
            return SearchResultsSchema(search_metadata=search_response.metadata)

        next_page_cursor = encode_next_page_cursor(search_response, page_cursor, query)
        has_next_page = next_page_cursor is not None
        page_info = PageInfoSchema(
            has_next_page=has_next_page,
            next_page_cursor=next_page_cursor,
        )

        cursor_info = None
        if search_response.total_items:
            cursor_info = CursorInfoSchema(
                total_items=search_response.total_items,
                start_cursor=search_response.start_cursor,
                end_cursor=search_response.end_cursor,
            )

        return SearchResultsSchema(
            data=search_response.results,
            page_info=page_info,
            search_metadata=search_response.metadata,
            cursor=cursor_info,
        )
    except (InvalidCursorError, ValueError) as e:
        raise HTTPException(status_code=status.HTTP_422_UNPROCESSABLE_ENTITY, detail=str(e))
    except QueryStateNotFoundError as e:
        raise HTTPException(status_code=status.HTTP_404_NOT_FOUND, detail=str(e))
    except Exception as e:
        raise HTTPException(
            status_code=status.HTTP_500_INTERNAL_SERVER_ERROR,
            detail=f"Search failed: {str(e)}",
        )


@router.post("/subscriptions", response_model=SearchResultsSchema[SearchResult])
async def search_subscriptions(
    request: SearchRequest,
    cursor: str | None = None,
) -> SearchResultsSchema[SearchResult]:
    return await _perform_search_and_fetch(EntityType.SUBSCRIPTION, request, cursor)


@router.post("/workflows", response_model=SearchResultsSchema[SearchResult])
async def search_workflows(
    request: SearchRequest,
    cursor: str | None = None,
) -> SearchResultsSchema[SearchResult]:
    return await _perform_search_and_fetch(EntityType.WORKFLOW, request, cursor)


@router.post("/products", response_model=SearchResultsSchema[SearchResult])
async def search_products(
    request: SearchRequest,
    cursor: str | None = None,
) -> SearchResultsSchema[SearchResult]:
    return await _perform_search_and_fetch(EntityType.PRODUCT, request, cursor)


@router.post("/processes", response_model=SearchResultsSchema[SearchResult])
async def search_processes(
    request: SearchRequest,
    cursor: str | None = None,
) -> SearchResultsSchema[SearchResult]:
    return await _perform_search_and_fetch(EntityType.PROCESS, request, cursor)


@router.get(
    "/paths",
    response_model=PathsResponse,
    response_model_exclude_none=True,
)
async def list_paths(
    prefix: str = Query("", min_length=0),
    q: str | None = Query(None, description="Query for path suggestions"),
    entity_type: EntityType = Query(EntityType.SUBSCRIPTION),
    limit: int = Query(10, ge=1, le=10),
) -> PathsResponse:

    if prefix:
        lquery_pattern = create_path_autocomplete_lquery(prefix)

        if not is_lquery_syntactically_valid(lquery_pattern, db.session):
            raise HTTPException(
                status_code=status.HTTP_422_UNPROCESSABLE_ENTITY,
                detail=f"Prefix '{prefix}' creates an invalid search pattern.",
            )
    stmt = build_paths_query(entity_type=entity_type, prefix=prefix, q=q)
    stmt = stmt.limit(limit)
    rows = db.session.execute(stmt).all()

    leaves, components = process_path_rows(rows)
    return PathsResponse(leaves=leaves, components=components)


@router.get(
    "/definitions",
    response_model=dict[UIType, TypeDefinition],
    response_model_exclude_none=True,
)
async def get_definitions() -> dict[UIType, TypeDefinition]:
    """Provide a static definition of operators and schemas for each UI type."""
    return generate_definitions()


@router.get(
    "/queries/{query_id}",
    response_model=SearchResultsSchema[SearchResult],
    summary="Retrieve saved search results by query_id",
)
async def get_by_query_id(
    query_id: str,
    cursor: str | None = None,
) -> SearchResultsSchema[SearchResult]:
    """Retrieve and execute a saved search by query_id."""
    return await _perform_search_and_fetch(query_id=query_id, cursor=cursor)


@router.get(
    "/queries/{query_id}/export",
    summary="Export query results by query_id",
    response_model=ExportResponse,
)
async def export_by_query_id(query_id: str) -> ExportResponse:
    """Export search results using query_id.

    The query is retrieved from the database, re-executed, and results are returned
    as flattened records suitable for CSV download.

    Args:
        query_id: QueryTypes UUID

    Returns:
        ExportResponse containing 'page' with an array of flattened entity records.

    Raises:
        HTTPException: 404 if query not found, 400 if invalid data
    """
    try:
        # Load SelectQuery from the database (what gets saved during search)
        query_state = QueryState.load_from_id(query_id, SelectQuery)

        # Convert to ExportQuery with export-appropriate limit
        export_query = ExportQuery(
            entity_type=query_state.query.entity_type,
            filters=query_state.query.filters,
            query_text=query_state.query.query_text,
        )

        export_records = await engine.execute_export(export_query, db.session, query_state.query_embedding)
        return ExportResponse(page=export_records)
    except ValueError as e:
        raise HTTPException(status_code=status.HTTP_422_UNPROCESSABLE_ENTITY, detail=str(e))
    except QueryStateNotFoundError as e:
        raise HTTPException(status_code=status.HTTP_404_NOT_FOUND, detail=str(e))
    except Exception as e:
        logger.error(e)
        raise HTTPException(
            status_code=status.HTTP_500_INTERNAL_SERVER_ERROR,
            detail=f"Error executing export: {str(e)}",
        )



# Copyright 2019-2025 SURF, GÉANT.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from datetime import date, datetime
from typing import Annotated, Any, Literal

from pydantic import BaseModel, BeforeValidator, Field, model_validator
from sqlalchemy import TIMESTAMP, and_
from sqlalchemy import cast as sa_cast
from sqlalchemy.sql.elements import ColumnElement

from orchestrator.search.core.types import FilterOp, SQLAColumn


def _validate_date_string(v: Any) -> Any:
    if not isinstance(v, str):
        return v
    try:
        datetime.fromisoformat(v)
        return v
    except Exception as exc:
        raise ValueError("is not a valid ISO-8601 date or datetime string") from exc


DateValue = datetime | date | str
ValidatedDateValue = Annotated[DateValue, BeforeValidator(_validate_date_string)]


class DateRange(BaseModel):

    start: ValidatedDateValue
    end: ValidatedDateValue

    @model_validator(mode="after")
    def _order(self) -> "DateRange":
        to_datetime = datetime.fromisoformat(str(self.end))
        from_datetime = datetime.fromisoformat(str(self.start))
        if to_datetime <= from_datetime:
            raise ValueError("'to' must be after 'from'")
        return self


class DateValueFilter(BaseModel):
    """A filter that operates on a single date value."""

    op: Literal[FilterOp.EQ, FilterOp.NEQ, FilterOp.LT, FilterOp.LTE, FilterOp.GT, FilterOp.GTE]
    value: ValidatedDateValue

    def to_expression(self, column: SQLAColumn, path: str) -> ColumnElement[bool]:
        date_column = sa_cast(column, TIMESTAMP(timezone=True))
        match self.op:
            case FilterOp.EQ:
                return date_column == self.value
            case FilterOp.NEQ:
                return date_column != self.value
            case FilterOp.LT:
                return date_column < self.value
            case FilterOp.LTE:
                return date_column <= self.value
            case FilterOp.GT:
                return date_column > self.value
            case FilterOp.GTE:
                return date_column >= self.value


class DateRangeFilter(BaseModel):
    """A filter that operates on a range of dates."""

    op: Literal[FilterOp.BETWEEN]
    value: DateRange

    def to_expression(self, column: SQLAColumn, path: str) -> ColumnElement[bool]:
        date_column = sa_cast(column, TIMESTAMP(timezone=True))
        return and_(date_column >= self.value.start, date_column < self.value.end)


DateFilter = Annotated[DateValueFilter | DateRangeFilter, Field(discriminator="op")]



#!/usr/bin/env python3
"""The main application module.

This module contains the main `LLMOrchestratorCore` class for the `FastAPI` backend and
provides the ability to run the CLI with LLM features (search and/or agent).
"""
# Copyright 2019-2025 SURF
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from typing import TYPE_CHECKING, Any

import typer
from structlog import get_logger

from orchestrator.app import OrchestratorCore
from orchestrator.cli.main import app as cli_app
from orchestrator.llm_settings import LLMSettings, llm_settings

if TYPE_CHECKING:
    from pydantic_ai.models.openai import OpenAIChatModel

logger = get_logger(__name__)


class LLMOrchestratorCore(OrchestratorCore):
    def __init__(
        self,
        *args: Any,
        llm_settings: LLMSettings = llm_settings,
        agent_model: "OpenAIChatModel | str | None" = None,
        **kwargs: Any,
    ) -> None:
        """Initialize the `LLMOrchestratorCore` class.

        This class extends `OrchestratorCore` with LLM features (search and agent).
        It runs the search migration based on feature flags.

        Args:
            *args: All the normal arguments passed to the `OrchestratorCore` class.
            llm_settings: A class of settings for the LLM
            agent_model: Override the agent model (defaults to llm_settings.AGENT_MODEL)
            **kwargs: Additional arguments passed to the `OrchestratorCore` class.

        Returns:
            None
        """
        self.llm_settings = llm_settings
        self.agent_model = agent_model or llm_settings.AGENT_MODEL

        super().__init__(*args, **kwargs)

        # Run search migration if search or agent is enabled
        if self.llm_settings.SEARCH_ENABLED or self.llm_settings.AGENT_ENABLED:
            logger.info("Running search migration")
            try:
                from orchestrator.db import db
                from orchestrator.search.llm_migration import run_migration

                with db.engine.begin() as connection:
                    run_migration(connection)
            except ImportError as e:
                logger.error(
                    "Unable to run search migration. Please install search dependencies: "
                    "`pip install orchestrator-core[search]`",
                    error=str(e),
                )
                raise


main_typer_app = typer.Typer()
main_typer_app.add_typer(cli_app, name="orchestrator", help="The orchestrator CLI commands")

if __name__ == "__main__":
    main_typer_app()



# Copyright 2019-2025 SURF, GÉANT.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from .base import (
    Aggregation,
    AggregationType,
    BaseAggregation,
    CountAggregation,
    FieldAggregation,
    TemporalGrouping,
    TemporalPeriod,
)

__all__ = [
    "Aggregation",
    "AggregationType",
    "BaseAggregation",
    "CountAggregation",
    "FieldAggregation",
    "TemporalGrouping",
    "TemporalPeriod",
]



# Copyright 2019-2025 SURF, GÉANT.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from functools import cache
from typing import Annotated

from fastapi import APIRouter, Depends, Request
from pydantic_ai.ag_ui import StateDeps, handle_ag_ui_request
from pydantic_ai.agent import Agent
from starlette.responses import Response
from structlog import get_logger

from orchestrator.search.agent import build_agent_instance
from orchestrator.search.agent.state import SearchState

router = APIRouter()
logger = get_logger(__name__)


@cache
def get_agent(request: Request) -> Agent[StateDeps[SearchState], str]:
    """Dependency to provide the agent instance.

    The agent is built once and cached for the lifetime of the application.
    """
    return build_agent_instance(request.app.agent_model)


@router.post("/")
async def agent_conversation(
    request: Request,
    agent: Annotated[Agent[StateDeps[SearchState], str], Depends(get_agent)],
) -> Response:
    """Agent conversation endpoint using pydantic-ai ag_ui protocol.

    This endpoint handles the interactive agent conversation for search.
    """
    initial_state = SearchState()
    return await handle_ag_ui_request(agent, request, deps=StateDeps(initial_state))



# Copyright 2019-2025 SURF, GÉANT.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from pydantic import BaseModel, ConfigDict, Field

from orchestrator.search.core.types import EntityType, RetrieverType
from orchestrator.search.filters import FilterTree
from orchestrator.search.query.queries import SelectQuery


class SearchRequest(BaseModel):
    """API request model for search operations.

    Only supports SELECT action, used by search endpoints.
    """

    filters: FilterTree | None = Field(
        default=None,
        description="Structured filters to apply to the search.",
    )
    query: str | None = Field(
        default=None,
        description="Text search query for semantic/fuzzy search.",
    )
    limit: int = Field(
        default=SelectQuery.DEFAULT_LIMIT,
        ge=SelectQuery.MIN_LIMIT,
        le=SelectQuery.MAX_LIMIT,
        description="Maximum number of search results to return.",
    )
    retriever: RetrieverType | None = Field(
        default=None,
        description="Force a specific retriever type. If None, uses default routing logic.",
    )

    model_config = ConfigDict(extra="forbid")

    def to_query(self, entity_type: EntityType) -> SelectQuery:
        """Convert API request to SelectQuery domain model.

        Args:
            entity_type: The entity type to search (provided by endpoint)

        Returns:
            SelectQuery for search operation
        """
        return SelectQuery(
            entity_type=entity_type,
            filters=self.filters,
            query_text=self.query,
            limit=self.limit,
            retriever=self.retriever,
        )



# Copyright 2019-2025 SURF.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from http import HTTPStatus

import structlog
from fastapi.routing import APIRouter

from orchestrator.schedules.service import add_scheduled_task_to_queue
from orchestrator.schemas.schedules import APSchedulerJobCreate, APSchedulerJobDelete, APSchedulerJobUpdate

logger = structlog.get_logger(__name__)

router: APIRouter = APIRouter()


@router.post("/", status_code=HTTPStatus.CREATED)
def create_scheduled_task(payload: APSchedulerJobCreate) -> dict[str, str]:
    """Create a scheduled task."""
    add_scheduled_task_to_queue(payload)
    return {"message": "Added to Create Queue", "status": "CREATED"}


@router.put("/", status_code=HTTPStatus.OK)
async def update_scheduled_task(payload: APSchedulerJobUpdate) -> dict[str, str]:
    """Update a scheduled task."""
    add_scheduled_task_to_queue(payload)
    return {"message": "Added to Update Queue", "status": "UPDATED"}


@router.delete("/", status_code=HTTPStatus.OK)
async def delete_scheduled_task(payload: APSchedulerJobDelete) -> dict[str, str]:
    """Delete a scheduled task."""
    add_scheduled_task_to_queue(payload)
    return {"message": "Added to Delete Queue", "status": "DELETED"}



# Copyright 2019-2025 SURF, GÉANT.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from collections import defaultdict
from typing import Any, Sequence

from pydantic import BaseModel, ConfigDict
from sqlalchemy import Select, String, case, cast, func, select
from sqlalchemy.engine import Row
from sqlalchemy.sql.elements import Label
from sqlalchemy.sql.selectable import CTE
from sqlalchemy_utils.types.ltree import Ltree

from orchestrator.db.models import AiSearchIndex
from orchestrator.search.aggregations import AggregationType, BaseAggregation, CountAggregation
from orchestrator.search.core.types import EntityType, FieldType, FilterOp, UIType
from orchestrator.search.filters import LtreeFilter
from orchestrator.search.query.mixins import OrderDirection
from orchestrator.search.query.queries import AggregateQuery, CountQuery, Query


class LeafInfo(BaseModel):
    """Information about a leaf (terminal field) in the entity schema."""

    name: str
    ui_types: list[UIType]
    paths: list[str]

    model_config = ConfigDict(
        extra="forbid",
        use_enum_values=True,
    )


class ComponentInfo(BaseModel):
    """Information about a component (nested object) in the entity schema."""

    name: str
    ui_types: list[UIType]

    model_config = ConfigDict(
        extra="forbid",
        use_enum_values=True,
    )


def create_path_autocomplete_lquery(prefix: str) -> str:
    """Create the lquery pattern for a multi-level path autocomplete search."""
    return f"{prefix}*.*"


def build_candidate_query(query: Query) -> Select:
    """Build the base query for retrieving candidate entities.

    Constructs a `SELECT` statement that retrieves distinct `entity_id` values
    from the index table for the given entity type, applying any structured
    filters from the provided query plan.

    Args:
        query: Any query type (SelectQuery, CountQuery, AggregateQuery) containing entity type and optional filters.

    Returns:
        Select: The SQLAlchemy `Select` object representing the query.
    """

    stmt = (
        select(AiSearchIndex.entity_id, AiSearchIndex.entity_title)
        .where(AiSearchIndex.entity_type == query.entity_type.value)
        .distinct()
    )

    if query.filters is not None:
        entity_id_col = AiSearchIndex.entity_id
        stmt = stmt.where(
            query.filters.to_expression(
                entity_id_col,
                entity_type_value=query.entity_type.value,
            )
        )

    return stmt


def build_paths_query(entity_type: EntityType, prefix: str | None = None, q: str | None = None) -> Select:
    """Build the query for retrieving paths and their value types for leaves/components processing."""
    stmt = select(AiSearchIndex.path, AiSearchIndex.value_type).where(AiSearchIndex.entity_type == entity_type.value)

    if prefix:
        lquery_pattern = create_path_autocomplete_lquery(prefix)
        ltree_filter = LtreeFilter(op=FilterOp.MATCHES_LQUERY, value=lquery_pattern)
        stmt = stmt.where(ltree_filter.to_expression(AiSearchIndex.path, path=""))

    stmt = stmt.group_by(AiSearchIndex.path, AiSearchIndex.value_type)

    if q:
        score = func.similarity(cast(AiSearchIndex.path, String), q)
        stmt = stmt.order_by(score.desc(), AiSearchIndex.path)
    else:
        stmt = stmt.order_by(AiSearchIndex.path)

    return stmt


def process_path_rows(rows: Sequence[Row]) -> tuple[list[LeafInfo], list[ComponentInfo]]:
    """Process query results to extract leaves and components information.

    Parameters
    ----------
    rows : Sequence[Row]
        Database rows containing path and value_type information

    Returns:
    -------
    tuple[list[LeafInfo], list[ComponentInfo]]
        Processed leaves and components
    """
    leaves_dict: dict[str, set[UIType]] = defaultdict(set)
    leaves_paths_dict: dict[str, set[str]] = defaultdict(set)
    components_set: set[str] = set()

    for row in rows:
        path, value_type = row

        path_str = str(path)
        path_segments = path_str.split(".")

        # Remove numeric segments
        clean_segments = [seg for seg in path_segments if not seg.isdigit()]

        if clean_segments:
            # Last segment is a leaf
            leaf_name = clean_segments[-1]
            ui_type = UIType.from_field_type(FieldType(value_type))
            leaves_dict[leaf_name].add(ui_type)
            leaves_paths_dict[leaf_name].add(path_str)

            # All segments except the first/last are components
            for component in clean_segments[1:-1]:
                components_set.add(component)

    leaves = [
        LeafInfo(name=leaf, ui_types=list(types), paths=sorted(leaves_paths_dict[leaf]))
        for leaf, types in leaves_dict.items()
    ]
    components = [ComponentInfo(name=component, ui_types=[UIType.COMPONENT]) for component in sorted(components_set)]

    return leaves, components


def _build_pivot_cte(base_query: Select, pivot_fields: list[str]) -> CTE:
    """Build CTE that pivots EAV rows into columns using CASE WHEN."""
    from orchestrator.search.aggregations import BaseAggregation

    pivot_columns = [AiSearchIndex.entity_id.label("entity_id")]

    for field_path in pivot_fields:
        pivot_columns.append(
            func.max(case((AiSearchIndex.path == Ltree(field_path), AiSearchIndex.value), else_=None)).label(
                BaseAggregation.field_to_alias(field_path)
            )
        )

    return (
        select(*pivot_columns)
        .where(
            AiSearchIndex.entity_id.in_(select(base_query.c.entity_id)),
            AiSearchIndex.path.in_([Ltree(p) for p in pivot_fields]),
        )
        .group_by(AiSearchIndex.entity_id)
        .cte("pivoted_entities")
    )


def _build_grouping_columns(
    query: CountQuery | AggregateQuery,
    pivot_cte: CTE,
) -> tuple[list[Any], list[Any], list[str]]:
    """Build GROUP BY columns and their SELECT columns.

    Args:
        query: CountQuery or AggregateQuery with group_by and temporal_group_by fields
        pivot_cte: The pivoted CTE containing entity fields as columns

    Returns:
        tuple: (select_columns, group_by_columns, group_column_names)
            - select_columns: List of labeled columns for SELECT
            - group_by_columns: List of columns for GROUP BY clause
            - group_column_names: List of column names (labels) that are grouping columns
    """

    select_columns = []
    group_by_columns = []
    group_column_names = []

    if query.group_by:
        for group_field in query.group_by:
            field_alias = BaseAggregation.field_to_alias(group_field)
            col = getattr(pivot_cte.c, field_alias)
            select_columns.append(col.label(field_alias))
            group_by_columns.append(col)
            group_column_names.append(field_alias)

    if query.temporal_group_by:
        for temp_group in query.temporal_group_by:
            select_col, group_col, col_name = temp_group.to_expression(pivot_cte.c)
            select_columns.append(select_col)
            group_by_columns.append(group_col)
            group_column_names.append(col_name)

    return select_columns, group_by_columns, group_column_names


def _build_aggregation_columns(query: CountQuery | AggregateQuery, pivot_cte: CTE) -> list[Label]:
    """Build aggregation columns (COUNT, SUM, AVG, MIN, MAX).

    Args:
        query: CountQuery or AggregateQuery
        pivot_cte: The pivoted CTE containing entity fields as columns

    Returns:
        List of labeled aggregation expressions
    """

    if isinstance(query, AggregateQuery):
        # AGGREGATE query with custom aggregations
        agg_columns = []
        for agg in query.aggregations:
            if isinstance(agg, CountAggregation):
                agg_columns.append(agg.to_expression(pivot_cte.c.entity_id))
            else:
                agg_columns.append(agg.to_expression(pivot_cte.c))
        return agg_columns

    # CountQuery without aggregations
    count_agg = CountAggregation(type=AggregationType.COUNT, alias="count")
    return [count_agg.to_expression(pivot_cte.c.entity_id)]


def _apply_cumulative_aggregations(
    stmt: Select,
    query: CountQuery | AggregateQuery,
    group_column_names: list[str],
    aggregation_columns: list[Label],
) -> Select:
    """Add cumulative aggregation columns."""

    # At this point, cumulative validation has already happened at query build time
    # in GroupingMixin.validate_grouping_constraints, so we know:
    # temporal_group_by exists and has exactly 1 element when cumulative=True
    if not query.cumulative or not aggregation_columns or not query.temporal_group_by:
        return stmt

    temporal_alias = query.temporal_group_by[0].alias

    base_subquery = stmt.subquery()
    partition_cols = [base_subquery.c[name] for name in group_column_names if name != temporal_alias]
    order_col = base_subquery.c[temporal_alias]

    base_columns = [base_subquery.c[col] for col in base_subquery.c.keys()]

    cumulative_columns = []
    for agg_col in aggregation_columns:
        cumulative_alias = f"{agg_col.key}_cumulative"
        over_kwargs: dict[str, Any] = {"order_by": order_col}
        if partition_cols:
            over_kwargs["partition_by"] = partition_cols
        cumulative_expr = func.sum(base_subquery.c[agg_col.key]).over(**over_kwargs).label(cumulative_alias)
        cumulative_columns.append(cumulative_expr)

    return select(*(base_columns + cumulative_columns)).select_from(base_subquery)


def _apply_ordering(
    stmt: Select,
    query: CountQuery | AggregateQuery,
    group_column_names: list[str],
) -> Select:
    """Apply ordering instructions to the SELECT statement."""
    columns_by_key = {col.key: col for col in stmt.selected_columns}

    if query.order_by:
        order_expressions = []
        for instruction in query.order_by:
            # 1) exact match
            col = columns_by_key.get(instruction.field)
            if col is None:
                # 2) temporal alias,
                for tg in query.temporal_group_by or []:
                    if instruction.field == tg.field or instruction.field == tg.alias:
                        col = columns_by_key.get(tg.alias)
                        if col is not None:
                            break
                if col is None:
                    # 3) normalized field path
                    col = columns_by_key.get(BaseAggregation.field_to_alias(instruction.field))
            if col is None:
                raise ValueError(f"Cannot order by '{instruction.field}'; column not found.")
            order_expressions.append(col.desc() if instruction.direction == OrderDirection.DESC else col.asc())
        return stmt.order_by(*order_expressions)

    if query.temporal_group_by:
        # Default ordering by all grouping columns (ascending)
        order_expressions = [columns_by_key[col_name].asc() for col_name in group_column_names]
        return stmt.order_by(*order_expressions)

    return stmt


def build_simple_count_query(base_query: Select) -> Select:
    """Build a simple count query without grouping.

    Args:
        base_query: Base candidate query with filters applied

    Returns:
        Select statement that counts distinct entity IDs
    """
    return select(func.count(func.distinct(base_query.c.entity_id)).label("total_count")).select_from(
        base_query.subquery()
    )


def build_aggregation_query(query: CountQuery | AggregateQuery, base_query: Select) -> tuple[Select, list[str]]:
    """Build aggregation query with GROUP BY and aggregation functions.

    Handles EAV storage by pivoting rows to columns, then applying SQL aggregations.
    This function only handles grouped aggregations. Simple counts are handled directly
    in the engine.

    Args:
        query: CountQuery or AggregateQuery with group_by and optional aggregations
        base_query: Base candidate query with filters applied

    Returns:
        tuple: (query_stmt, group_column_names)
            - query_stmt: SQLAlchemy Select statement for grouped aggregation
            - group_column_names: List of column names that are grouping columns
    """
    pivot_cte = _build_pivot_cte(base_query, query.get_pivot_fields())
    select_cols, group_cols, group_col_names = _build_grouping_columns(query, pivot_cte)
    agg_cols = _build_aggregation_columns(query, pivot_cte)

    stmt = select(*(select_cols + agg_cols)).select_from(pivot_cte)
    if group_cols:
        stmt = stmt.group_by(*group_cols)

    stmt = _apply_cumulative_aggregations(stmt, query, group_col_names, agg_cols)
    stmt = _apply_ordering(stmt, query, group_col_names)

    return stmt, group_col_names



# Copyright 2019-2023 SURF.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import uuid
from datetime import datetime
from typing import Any, Callable

import pytz
import sqlalchemy
from dateutil.parser import parse
from sqlalchemy import BinaryExpression, Cast, ColumnClause, ColumnElement, String, cast
from sqlalchemy.sql.functions import coalesce
from sqlalchemy.sql.operators import eq

from orchestrator.settings import app_settings
from orchestrator.utils.search_query import Node, WhereCondGenerator


def convert_to_datetime(value: str) -> datetime:
    """Parse iso 8601 date from string to datetime.

    Example date: "2022-07-21T03:40:48+00:00"
    """
    try:
        return parse(value).replace(tzinfo=pytz.UTC)
    except ValueError:
        raise ValueError(f"{value} is not a valid date")


def convert_to_int(value: str) -> int:
    try:
        return int(value)
    except ValueError:
        raise ValueError(f"{value} is not a valid integer")


def _phrase_to_ilike_str(phrase_node: Node) -> str:
    def to_str(node: Node) -> str:
        node_type, value = node
        return f"{value}%" if node_type == "PrefixWord" else f"{value}"

    return " ".join(to_str(w) for w in phrase_node[1])


def _coalesce_if_nullable(field: ColumnElement, default_value: Any = "") -> ColumnElement:
    if isinstance(field, Cast):
        is_nullable = field.wrapped_column_expression.nullable
    else:
        is_nullable = getattr(field, "nullable", False)
    return coalesce(field, default_value) if is_nullable else field


def _filter_string(field: ColumnElement) -> WhereCondGenerator:
    field = _coalesce_if_nullable(field)

    def _clause_gen(node: Node) -> BinaryExpression:
        if node[0] == "Phrase":
            return field.ilike(_phrase_to_ilike_str(node))
        if node[0] == "ValueGroup":
            vals = [w[1] for w in node[1] if w[0] in ["Word", "PrefixWord"]]  # Only works for (Prefix)Words atm
            return field.in_(vals)
        return field.ilike(f"{node[1]}") if app_settings.FILTER_BY_MODE == "exact" else field.ilike(f"%{node[1]}%")

    return _clause_gen


def _filter_as_string(field: ColumnClause) -> WhereCondGenerator:
    return _filter_string(cast(field, String))


RANGE_TYPES = {
    ">": ColumnClause.__gt__,
    ">=": ColumnClause.__ge__,
    "<": ColumnClause.__lt__,
    "<=": ColumnClause.__le__,
    "!": ColumnClause.__ne__,
}


def _filter_comparable(field: ColumnClause, value_converter: Callable[[str], Any]) -> WhereCondGenerator:
    _default_clause = _filter_as_string(field)

    def _clause_gen(node: Node) -> ColumnElement[bool]:
        v = node_to_str_val(node)
        range_comparison_op = next((op for op in [">=", "<=", ">", "<"] if v.startswith(op)), "")
        if range_comparison_op:
            op = RANGE_TYPES.get(range_comparison_op)
            return op(field, value_converter(v[len(range_comparison_op) :]))  # type:ignore
        return _default_clause(node)

    return _clause_gen


def _value_as_bool(v: str) -> bool | None:
    lower = v.lower()
    if lower == "true":
        return True
    if lower == "false":
        return False
    return None


def _filter_bool(field: ColumnClause) -> WhereCondGenerator:
    def _clause_gen(node: Node) -> BinaryExpression | ColumnElement[bool]:
        if node[0] in ["Phrase", "ValueGroup"]:
            vals = [
                boolean_val for w in node[1] if (boolean_val := _value_as_bool(w[1]))
            ]  # Only works for (Prefix)Words atm
            return field.in_(vals)
        boolean_val = _value_as_bool(node[1])
        return eq(field, boolean_val) if boolean_val is not None else sqlalchemy.false()

    return _clause_gen


def inferred_filter(field: ColumnClause) -> WhereCondGenerator:
    python_type = field.type.python_type
    if python_type is str:
        return _filter_string(field)
    if python_type is uuid.UUID:
        return _filter_as_string(field)
    if python_type is bool:
        return _filter_bool(field)
    if python_type is datetime:
        return _filter_comparable(field, convert_to_datetime)
    if python_type is int:
        return _filter_comparable(field, convert_to_int)

    raise Exception(f"Unsupported column type for generic filter: {field}")


def node_to_str_val(node: Node, *, sep: str = " ") -> str:
    if node[0] in ["Phrase", "ValueGroup"]:
        return sep.join(w[1] for w in node[1])
    return node[1]


def filter_exact(field: ColumnClause) -> WhereCondGenerator:
    def _clause_gen(node: Node) -> BinaryExpression[bool] | ColumnElement[bool]:
        if node[0] == "Word":
            return field.ilike(f"{node[1]}")
        if node[0] == "PrefixWord":
            return field.ilike(f"{node[1]}%")
        return _filter_string(field)(node)

    return _clause_gen


def filter_uuid_exact(field: ColumnClause) -> WhereCondGenerator:
    assert field.type.python_type == uuid.UUID  # noqa: S101

    def _clause_gen(node: Node) -> BinaryExpression[bool] | ColumnElement[bool]:
        try:
            v = uuid.UUID(node_to_str_val(node))
        except ValueError:
            # Not a valid uuid
            return sqlalchemy.false()
        return eq(field, v)

    return _clause_gen



from typing import Annotated, Union

import strawberry
from strawberry.scalars import JSON

from orchestrator.graphql.schemas.errors import Error
from orchestrator.schemas import WorkerStatus
from orchestrator.schemas.engine_settings import EngineSettingsSchema

CACHE_FLUSH_OPTIONS: dict[str, str] = {"all": "All caches"}


@strawberry.experimental.pydantic.type(model=WorkerStatus, all_fields=True)
class WorkerStatusType:
    pass


@strawberry.experimental.pydantic.type(model=EngineSettingsSchema, all_fields=True)
class EngineSettingsType:
    pass


@strawberry.type
class StatusType:
    engine_settings: EngineSettingsType | None
    worker_status: WorkerStatusType | None
    cache_names: JSON | None


# Responses
@strawberry.type
class CacheClearSuccess:
    deleted: int


CacheClearResponse = Annotated[Union[CacheClearSuccess, Error], strawberry.union("CacheClearResponse")]
StatusUpdateResponse = Annotated[Union[EngineSettingsType, Error], strawberry.union("StatusUpdateResponse")]



# Copyright 2019-2025 SURF, GÉANT.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Any

import structlog
from pydantic_ai.ag_ui import StateDeps
from pydantic_ai.agent import Agent
from pydantic_ai.models.openai import OpenAIChatModel
from pydantic_ai.settings import ModelSettings
from pydantic_ai.toolsets import FunctionToolset

from orchestrator.search.agent.prompts import get_base_instructions, get_dynamic_instructions
from orchestrator.search.agent.state import SearchState
from orchestrator.search.agent.tools import search_toolset

logger = structlog.get_logger(__name__)


def build_agent_instance(
    model: str | OpenAIChatModel, agent_tools: list[FunctionToolset[Any]] | None = None
) -> Agent[StateDeps[SearchState], str]:
    """Build and configure the search agent instance.

    Args:
        model: The LLM model to use (string or OpenAIChatModel instance)
        agent_tools: Optional list of additional toolsets to include

    Returns:
        Configured Agent instance with StateDeps[SearchState] dependencies

    Raises:
        Exception: If agent initialization fails
    """
    toolsets = agent_tools + [search_toolset] if agent_tools else [search_toolset]

    agent = Agent(
        model=model,
        deps_type=StateDeps[SearchState],
        model_settings=ModelSettings(
            parallel_tool_calls=False,
        ),  # https://github.com/pydantic/pydantic-ai/issues/562
        toolsets=toolsets,
    )
    agent.instructions(get_base_instructions)
    agent.instructions(get_dynamic_instructions)

    return agent



# Copyright 2019-2020 SURF.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from typing import cast
from uuid import UUID

import structlog
from graphql import GraphQLError
from pydantic.alias_generators import to_camel as to_lower_camel
from sqlalchemy import Select, func, select
from sqlalchemy.orm import contains_eager
from starlette.concurrency import run_in_threadpool
from strawberry.experimental.pydantic.conversion_types import StrawberryTypeFromPydantic

from nwastdlib.asyncio import gather_nice
from orchestrator.db import ProductTable, SubscriptionTable, db
from orchestrator.db.filters import Filter
from orchestrator.db.filters.subscription import (
    filter_by_query_string,
    filter_subscriptions,
    subscription_filter_fields,
)
from orchestrator.db.range import apply_range_to_statement
from orchestrator.db.sorting import Sort
from orchestrator.db.sorting.subscription import (
    sort_subscriptions,
    subscription_sort_fields,
)
from orchestrator.graphql.pagination import Connection
from orchestrator.graphql.schemas.product import ProductModelGraphql
from orchestrator.graphql.schemas.subscription import SubscriptionInterface
from orchestrator.graphql.types import (
    GraphqlFilter,
    GraphqlSort,
    OrchestratorInfo,
)
from orchestrator.graphql.utils import (
    create_resolver_error_handler,
    is_query_detailed,
    is_querying_page_data,
    to_graphql_result_page,
)
from orchestrator.utils.get_subscription_dict import get_subscription_dict

logger = structlog.get_logger(__name__)
# Note: we can make this more fancy by adding metadata to the field annotation that indicates if a resolver
# needs subscription details or not, and use that information here. Left as an exercise for the reader.
base_sub_props = tuple(
    [to_lower_camel(key) for key in SubscriptionInterface.__annotations__]
    + [to_lower_camel(key) for key in ProductModelGraphql.__annotations__]
    + ["__typename"]
)

_is_subscription_detailed = is_query_detailed(base_sub_props, ("SubscriptionInterface",))


def get_subscription_graphql_type(info: OrchestratorInfo, subscription_name: str) -> StrawberryTypeFromPydantic:
    subscription_graphql_type = info.context.graphql_models.get(subscription_name)
    if not subscription_graphql_type:
        logger.warning(message=f"No graphql type found for {subscription_name}")
        base_type = info.context.graphql_models.get("subscription")
        if not base_type:
            raise GraphQLError("No subscription base type found")
        return base_type
    return subscription_graphql_type


async def get_subscription_details(info: OrchestratorInfo, subscription: SubscriptionTable) -> SubscriptionInterface:
    from orchestrator.domain import SUBSCRIPTION_MODEL_REGISTRY
    from orchestrator.graphql.autoregistration import graphql_subscription_name

    subscription_dict_data, _ = await get_subscription_dict(subscription.subscription_id)

    domain_model_type = SUBSCRIPTION_MODEL_REGISTRY[subscription.product.name]
    base_model = domain_model_type.__base_type__ or domain_model_type

    subscription_name = graphql_subscription_name(base_model.__name__)
    subscription_details = base_model.model_validate(subscription_dict_data, strict=False)
    subscription_details._db_model = subscription  # type: ignore

    strawberry_type = get_subscription_graphql_type(info, subscription_name)
    return strawberry_type.from_pydantic(subscription_details)  # type: ignore


async def format_subscription(info: OrchestratorInfo, subscription: SubscriptionTable) -> SubscriptionInterface:
    if _is_subscription_detailed(info):
        return await get_subscription_details(info, subscription)

    strawberry_type = get_subscription_graphql_type(info, "subscription")
    return strawberry_type.from_pydantic(subscription)  # type:ignore


async def resolve_subscription(info: OrchestratorInfo, id: UUID) -> SubscriptionInterface | None:
    stmt = select(SubscriptionTable).where(SubscriptionTable.subscription_id == id)

    if subscription := await run_in_threadpool(db.session.scalar, stmt):
        return await format_subscription(info, subscription)
    return None


async def resolve_subscriptions(
    info: OrchestratorInfo,
    filter_by: list[GraphqlFilter] | None = None,
    sort_by: list[GraphqlSort] | None = None,
    first: int = 10,
    after: int = 0,
    query: str | None = None,
) -> Connection[SubscriptionInterface]:
    _error_handler = create_resolver_error_handler(info)

    pydantic_filter_by: list[Filter] = [item.to_pydantic() for item in filter_by] if filter_by else []
    pydantic_sort_by: list[Sort] = [item.to_pydantic() for item in sort_by] if sort_by else []
    logger.debug(
        "resolve_subscription() called",
        range=[after, after + first],
        sort=sort_by,
        filter=pydantic_filter_by,
        query=query,
    )

    stmt = (
        select(SubscriptionTable)
        .join(ProductTable)
        .options(
            # contains_eager() is needed because .join() does not eagerload, unlike options(joinedload())
            # (and using joinedload() is not possible because of filter_subscriptions())
            contains_eager(SubscriptionTable.product),
        )
    )

    stmt = filter_subscriptions(stmt, pydantic_filter_by, _error_handler)
    if query is not None:
        stmt = filter_by_query_string(stmt, query)

    stmt = cast(Select, sort_subscriptions(stmt, pydantic_sort_by, _error_handler))
    total = await run_in_threadpool(db.session.scalar, select(func.count()).select_from(stmt.subquery()))
    stmt = apply_range_to_statement(stmt, after, after + first + 1)

    graphql_subscriptions: list[SubscriptionInterface] = []
    if is_querying_page_data(info):
        scalars = await run_in_threadpool(db.session.scalars, stmt)
        subscriptions = scalars.all()
        graphql_subscriptions = list(await gather_nice((format_subscription(info, p) for p in subscriptions)))  # type: ignore
    logger.info("Resolve subscriptions", filter_by=filter_by, total=total)

    return to_graphql_result_page(
        graphql_subscriptions, first, after, total, subscription_sort_fields(), subscription_filter_fields()
    )



import structlog
import typer
from sqlalchemy import text
from sqlalchemy.exc import SQLAlchemyError

from orchestrator.db import db
from orchestrator.db.models import AiSearchIndex, SearchQueryTable
from orchestrator.llm_settings import llm_settings

logger = structlog.get_logger(__name__)

app = typer.Typer(
    name="embedding",
    help="Resize vector dimensions of the embeddings.",
)


def get_current_embedding_dimension() -> int | None:
    """Get the current dimension of the embedding column from ai_search_index table.

    Returns:
        Current dimension size or None if no records exist or column doesn't exist
    """
    try:
        query = text(
            """
            SELECT vector_dims(embedding) as dimension
            FROM ai_search_index
            WHERE embedding IS NOT NULL
            LIMIT 1
        """
        )
        result = db.session.execute(query).fetchone()
        if result and result[0]:
            return result[0]
        return None

    except SQLAlchemyError as e:
        logger.error("Failed to get current embedding dimension", error=str(e))
        return None


def drop_all_embeddings() -> tuple[int, int]:
    """Drop all records from ai_search_index and search_queries tables.

    Returns:
        Tuple of (ai_search_index records deleted, search_queries records deleted)
    """
    try:
        index_deleted = db.session.query(AiSearchIndex).delete()
        query_deleted = db.session.query(SearchQueryTable).delete()
        db.session.commit()
        logger.info(
            f"Deleted {index_deleted} records from ai_search_index and {query_deleted} records from search_queries"
        )
        return index_deleted, query_deleted

    except SQLAlchemyError as e:
        db.session.rollback()
        logger.error("Failed to drop embeddings records", error=str(e))
        raise


def alter_embedding_column_dimension(new_dimension: int) -> None:
    """Alter the embedding columns in both ai_search_index and search_queries tables.

    Args:
        new_dimension: New vector dimension size
    """
    try:
        db.session.execute(text("ALTER TABLE ai_search_index DROP COLUMN IF EXISTS embedding"))
        db.session.execute(text(f"ALTER TABLE ai_search_index ADD COLUMN embedding vector({new_dimension})"))

        db.session.execute(text("ALTER TABLE search_queries DROP COLUMN IF EXISTS query_embedding"))
        db.session.execute(text(f"ALTER TABLE search_queries ADD COLUMN query_embedding vector({new_dimension})"))

        db.session.commit()

        db.session.close()

        logger.info(f"Altered embedding columns to dimension {new_dimension} in ai_search_index and search_queries")

    except SQLAlchemyError as e:
        db.session.rollback()
        logger.error("Failed to alter embedding column dimensions", error=str(e))
        raise


@app.command("resize")
def resize_embeddings_command() -> None:
    """Resize vector dimensions of embedding columns in ai_search_index and search_queries tables.

    Compares the current embedding dimension in the database with the configured
    dimension in llm_settings. If they differ, drops all records and alters both
    embedding columns to match the new dimension.
    """
    new_dimension = llm_settings.EMBEDDING_DIMENSION

    logger.info("Starting embedding dimension resize", new_dimension=new_dimension)

    current_dimension = get_current_embedding_dimension()

    if current_dimension is None:
        logger.warning("Could not determine current dimension for embedding column")

    if current_dimension == new_dimension:
        logger.info(
            "Embedding dimensions match, no resize needed",
            current_dimension=current_dimension,
            new_dimension=new_dimension,
        )
        return

    logger.info("Dimension mismatch detected", current_dimension=current_dimension, new_dimension=new_dimension)

    if not typer.confirm(
        "This will DELETE ALL RECORDS from ai_search_index and search_queries tables and alter embedding columns. Continue?"
    ):
        logger.info("Operation cancelled by user")
        return

    try:
        # Drop all records first.
        logger.info("Dropping all embedding records...")
        index_deleted, query_deleted = drop_all_embeddings()

        # Then alter column dimensions.
        logger.info(f"Altering embedding columns to dimension {new_dimension}...")
        alter_embedding_column_dimension(new_dimension)

        logger.info(
            "Embedding dimension resize completed successfully",
            index_records_deleted=index_deleted,
            query_records_deleted=query_deleted,
            new_dimension=new_dimension,
        )

    except Exception as e:
        logger.error("Embedding dimension resize failed", error=str(e))
        raise typer.Exit(1)


if __name__ == "__main__":
    app()



# Copyright 2019-2020 SURF, GÉANT.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from datetime import timedelta

from sqlalchemy import select

from orchestrator.db import ProcessTable, db
from orchestrator.settings import app_settings, get_authorizers
from orchestrator.targets import Target
from orchestrator.utils.datetime import nowtz
from orchestrator.workflow import ProcessStatus, StepList, done, init, step, workflow
from pydantic_forms.types import State

authorizers = get_authorizers()


@step("Clean up completed tasks older than TASK_LOG_RETENTION_DAYS")
def remove_tasks() -> State:
    cutoff = nowtz() - timedelta(days=app_settings.TASK_LOG_RETENTION_DAYS)
    tasks = db.session.scalars(
        select(ProcessTable)
        .filter(ProcessTable.is_task.is_(True))
        .filter(ProcessTable.last_status == ProcessStatus.COMPLETED)
        .filter(ProcessTable.last_modified_at <= cutoff)
    )
    count = 0
    for task in tasks:
        db.session.delete(task)
        count += 1

    return {"tasks_removed": count}


@workflow(
    "Clean up old tasks",
    target=Target.SYSTEM,
    authorize_callback=authorizers.authorize_callback,
    retry_auth_callback=authorizers.retry_auth_callback,
)
def task_clean_up_tasks() -> StepList:
    return init >> remove_tasks >> done



# Copyright 2019-2025 SURF, GÉANT.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Any, Literal

from pydantic import BaseModel, Field


class JSONPatchOp(BaseModel):
    """A JSON Patch operation (RFC 6902).

    Docs reference: https://docs.ag-ui.com/concepts/state
    """

    op: Literal["add", "remove", "replace", "move", "copy", "test"] = Field(
        description="The operation to perform: add, remove, replace, move, copy, or test"
    )
    path: str = Field(description="JSON Pointer (RFC 6901) to the target location")
    value: Any | None = Field(
        default=None,
        description="The value to apply (for add, replace operations)",
    )
    from_: str | None = Field(
        default=None,
        alias="from",
        description="Source path (for move, copy operations)",
    )

    @classmethod
    def upsert(cls, path: str, value: Any, existed: bool) -> "JSONPatchOp":
        """Create an add or replace operation depending on whether the path existed.

        Args:
            path: JSON Pointer path to the target location
            value: The value to set
            existed: True if the path already exists (use replace), False otherwise (use add)

        Returns:
            JSONPatchOp with 'replace' if existed is True, 'add' otherwise
        """
        return cls(op="replace" if existed else "add", path=path, value=value)



# Copyright 2019-2025 SURF, GÉANT.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import structlog
from sqlalchemy import Select, func, select
from sqlalchemy.orm import Session

from orchestrator.search.core.embedding import QueryEmbedder
from orchestrator.search.core.types import EntityType, RetrieverType, SearchMetadata
from orchestrator.search.query.results import (
    AggregationResponse,
    SearchResponse,
    format_aggregation_response,
    format_search_response,
)
from orchestrator.search.retrieval.pagination import PageCursor
from orchestrator.search.retrieval.retrievers import (
    FuzzyRetriever,
    ProcessHybridRetriever,
    Retriever,
    RrfHybridRetriever,
    SemanticRetriever,
)
from orchestrator.search.retrieval.retrievers.structured import StructuredRetriever

from .builder import build_aggregation_query, build_candidate_query, build_simple_count_query
from .export import fetch_export_data
from .queries import AggregateQuery, CountQuery, ExportQuery, SelectQuery

logger = structlog.get_logger(__name__)


def _get_retriever_from_override(
    query: SelectQuery | ExportQuery,
    cursor: PageCursor | None,
    query_embedding: list[float] | None,
) -> Retriever | None:
    """Get retriever instance from explicit override, or None if no override.

    Args:
        query: Query that may have a retriever override
        cursor: Pagination cursor
        query_embedding: Pre-computed embedding (may be None)

    Returns:
        Retriever instance matching the requested type, or None if no override specified

    Raises:
        ValueError: If override requirements aren't met (e.g., no query text or embedding)
    """
    if query.retriever is None:
        return None

    retriever_type = query.retriever

    # Validate query_text (required for all retriever types)
    if not query.query_text:
        raise ValueError(f"{retriever_type.value.capitalize()} retriever requested but no query text provided.")

    is_process = query.entity_type == EntityType.PROCESS

    if retriever_type == RetrieverType.FUZZY:
        return (
            ProcessHybridRetriever(None, query.query_text, cursor)
            if is_process
            else FuzzyRetriever(query.query_text, cursor)
        )
    if retriever_type == RetrieverType.SEMANTIC:
        if query_embedding is None:
            raise ValueError(
                "Semantic retriever requested but query embedding is not available. "
                "Embedding generation may have failed."
            )
        return SemanticRetriever(query_embedding, cursor)
    if query_embedding is None:
        raise ValueError(
            "Hybrid retriever requested but query embedding is not available. " "Embedding generation may have failed."
        )
    return (
        ProcessHybridRetriever(query_embedding, query.query_text, cursor)
        if is_process
        else RrfHybridRetriever(query_embedding, query.query_text, cursor)
    )


def _create_cursor_info(
    final_stmt: Select,
    db_session: Session,
    cursor: PageCursor | None,
    query: SelectQuery | ExportQuery,
    query_embedding: list[float] | None,
    candidate_query: Select,
    row_count: int,
) -> tuple[int | None, int | None, int | None]:
    total_count_stmt = Retriever.route(query, None, query_embedding).apply(candidate_query) if cursor else final_stmt
    total_items_or_none = db_session.scalar(select(func.count()).select_from(total_count_stmt.subquery()))
    total_items = total_items_or_none or 0
    count_with_cursor_or_none = (
        db_session.scalar(select(func.count()).select_from(final_stmt.subquery())) if cursor else total_items
    )
    count_with_cursor = count_with_cursor_or_none or 0
    start_cursor = total_items - count_with_cursor
    row_end_cursor_count = row_count - (2 if row_count > query.limit else 1)
    end_cursor = start_cursor + row_end_cursor_count
    return total_items, start_cursor, end_cursor


async def _execute_search(
    query: SelectQuery | ExportQuery,
    db_session: Session,
    limit: int,
    cursor: PageCursor | None = None,
    query_embedding: list[float] | None = None,
) -> SearchResponse:
    """Internal implementation to execute search with specified query.

    Args:
        query: The SELECT or EXPORT query with vector, fuzzy, or filter criteria.
        db_session: The active SQLAlchemy session for executing the query.
        limit: Maximum number of results to return.
        cursor: Optional pagination cursor.
        query_embedding: Optional pre-computed query embedding to use instead of generating a new one.

    Returns:
        SearchResponse with results and embedding (for internal use).
    """
    if not query.vector_query and not query.filters and not query.fuzzy_term:
        logger.warning("No search criteria provided (vector_query, fuzzy_term, or filters).")
        return SearchResponse(results=[], metadata=SearchMetadata.empty())

    candidate_query = build_candidate_query(query)

    if query.vector_query and not query_embedding:
        query_embedding = await QueryEmbedder.generate_for_text_async(query.vector_query)

    # Get retriever (from override or automatic routing)
    retriever = _get_retriever_from_override(query, cursor, query_embedding) or Retriever.route(
        query, cursor, query_embedding
    )
    logger.debug("Using retriever", retriever_type=retriever.__class__.__name__)

    final_stmt = retriever.apply(candidate_query)
    final_stmt_with_limit = final_stmt.limit(limit)
    logger.debug(final_stmt_with_limit)

    result = db_session.execute(final_stmt_with_limit).mappings().all()
    result_rows = list(result)
    row_count = len(result_rows)

    total_items: int | None = None
    start_cursor: int | None = None
    end_cursor: int | None = None
    if isinstance(retriever, StructuredRetriever) and row_count > 0:
        total_items, start_cursor, end_cursor = _create_cursor_info(
            final_stmt, db_session, cursor, query, query_embedding, candidate_query, row_count
        )

    return format_search_response(
        result_rows, query, retriever.metadata, query_embedding, total_items, start_cursor, end_cursor
    )


async def execute_search(
    query: SelectQuery,
    db_session: Session,
    cursor: PageCursor | None = None,
    query_embedding: list[float] | None = None,
) -> SearchResponse:
    """Execute a SELECT search query.

    This executes a SELECT action search using vector/fuzzy/filter search with ranking.

    Args:
        query: SelectQuery with search criteria
        db_session: Database session
        cursor: Optional pagination cursor
        query_embedding: Optional pre-computed embedding

    Returns:
        SearchResponse with ranked results
    """

    # Fetch one extra to determine if there is a next page
    fetch_limit = query.limit + 1 if query.limit > 0 else query.limit
    response = await _execute_search(query, db_session, fetch_limit, cursor, query_embedding)
    has_more = len(response.results) > query.limit and query.limit > 0

    # Trim to requested limit
    response.results = response.results[: query.limit]
    response.has_more = has_more

    return response


async def execute_export(
    query: ExportQuery,
    db_session: Session,
    query_embedding: list[float] | None = None,
) -> list[dict]:
    """Execute a search and export flattened entity data.

    Args:
        query: ExportQuery with search criteria
        db_session: Database session
        query_embedding: Optional pre-computed embedding

    Returns:
        List of flattened entity records suitable for export.
    """
    search_response = await _execute_search(
        query=query,
        db_session=db_session,
        limit=query.limit,
        query_embedding=query_embedding,
    )

    entity_ids = [res.entity_id for res in search_response.results]
    return fetch_export_data(query.entity_type, entity_ids)


async def execute_aggregation(
    query: CountQuery | AggregateQuery,
    db_session: Session,
) -> AggregationResponse:
    """Execute aggregation query and return formatted results.

    Args:
        query: CountQuery or AggregateQuery
        db_session: Database session

    Returns:
        AggregationResponse with results and metadata
    """
    candidate_query = build_candidate_query(query)

    if isinstance(query, CountQuery) and not query.group_by and not query.temporal_group_by:
        # Simple count without grouping
        agg_query = build_simple_count_query(candidate_query)
        group_column_names: list[str] = []
    else:
        # Grouped aggregation - needs pivoting
        agg_query, group_column_names = build_aggregation_query(query, candidate_query)

    logger.debug("Executing aggregation query", sql=str(agg_query))

    result_rows = db_session.execute(agg_query).mappings().all()

    return format_aggregation_response(result_rows, group_column_names, query)



# Copyright 2019-2025 SURF, GÉANT.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from abc import abstractmethod
from enum import Enum
from typing import Annotated, Any, Literal, TypeAlias

from pydantic import BaseModel, ConfigDict, Field
from sqlalchemy import Integer, cast, func
from sqlalchemy.sql.elements import ColumnElement, Label


class AggregationType(str, Enum):
    """Types of aggregations that can be computed."""

    COUNT = "count"
    SUM = "sum"
    AVG = "avg"
    MIN = "min"
    MAX = "max"


class TemporalPeriod(str, Enum):
    """Time periods for temporal grouping."""

    YEAR = "year"
    QUARTER = "quarter"
    MONTH = "month"
    WEEK = "week"
    DAY = "day"
    HOUR = "hour"


class TemporalGrouping(BaseModel):
    """Defines temporal grouping for date/time fields.

    Used to group query results by time periods (e.g., group subscriptions by start_date per month).
    """

    field: str = Field(description="The datetime field path to group by temporally.")
    period: TemporalPeriod = Field(description="The time period to group by (year, quarter, month, week, day, hour).")

    model_config = ConfigDict(
        extra="forbid",
        json_schema_extra={
            "examples": [
                {"field": "subscription.start_date", "period": "month"},
                {"field": "subscription.end_date", "period": "year"},
                {"field": "process.created_at", "period": "day"},
            ]
        },
    )

    @property
    def alias(self) -> str:
        """Return the SQL-friendly alias for this temporal grouping."""
        return f"{BaseAggregation.field_to_alias(self.field)}_{self.period.value}"

    def get_pivot_fields(self) -> list[str]:
        """Return fields that need to be pivoted for this temporal grouping."""
        return [self.field]

    def to_expression(self, pivot_cte_columns: Any) -> tuple[Label, Any, str]:
        """Build SQLAlchemy expression for temporal grouping.

        Args:
            pivot_cte_columns: The columns object from the pivot CTE

        Returns:
            tuple: (select_column, group_by_column, column_name)
                - select_column: Labeled column for SELECT
                - group_by_column: Column expression for GROUP BY
                - column_name: The label/name of the column in results
        """
        from sqlalchemy import TIMESTAMP, cast, func

        field_alias = BaseAggregation.field_to_alias(self.field)
        col = getattr(pivot_cte_columns, field_alias)
        truncated_col = func.date_trunc(self.period.value, cast(col, TIMESTAMP(timezone=True)))

        col_name = self.alias
        select_col = truncated_col.label(col_name)
        return select_col, truncated_col, col_name


class BaseAggregation(BaseModel):
    """Base class for all aggregation types."""

    type: AggregationType = Field(description="The type of aggregation to perform.")
    alias: str = Field(description="The name for this aggregation in the results.")

    @classmethod
    def create(cls, data: dict) -> "Aggregation":
        """Create the correct aggregation instance based on type field.

        Args:
            data: Dictionary with aggregation data including 'type' discriminator

        Returns:
            Validated aggregation instance (CountAggregation or FieldAggregation)

        Raises:
            ValidationError: If data is invalid or type is unknown
        """
        from pydantic import TypeAdapter

        adapter: TypeAdapter = TypeAdapter(Aggregation)
        return adapter.validate_python(data)

    @staticmethod
    def field_to_alias(field_path: str) -> str:
        """Convert field path to SQL column alias.

        Examples:
            'subscription.name' -> 'subscription_name'
            'product.serial-number' -> 'product_serial_number'
        """
        return field_path.replace(".", "_").replace("-", "_")

    def get_pivot_fields(self) -> list[str]:
        """Return fields that need to be pivoted for this aggregation."""
        return []

    @abstractmethod
    def to_expression(self, *args: Any, **kwargs: Any) -> Label:
        """Build SQLAlchemy expression for this aggregation.

        Returns:
            Label: A labeled SQLAlchemy expression
        """
        raise NotImplementedError


class CountAggregation(BaseAggregation):
    """Count aggregation - counts number of entities."""

    type: Literal[AggregationType.COUNT]

    def to_expression(self, entity_id_column: ColumnElement) -> Label:
        """Build SQLAlchemy expression for count aggregation.

        Args:
            entity_id_column: The entity_id column from the pivot CTE

        Returns:
            Label: A labeled SQLAlchemy expression
        """
        return func.count(entity_id_column).label(self.alias)


class FieldAggregation(BaseAggregation):
    """Field-based aggregation (sum, avg, min, max)."""

    type: Literal[AggregationType.SUM, AggregationType.AVG, AggregationType.MIN, AggregationType.MAX]
    field: str = Field(description="The field path to aggregate on.")

    def get_pivot_fields(self) -> list[str]:
        """Return fields that need to be pivoted for this aggregation."""
        return [self.field]

    def to_expression(self, pivot_cte_columns: Any) -> Label:
        """Build SQLAlchemy expression for field-based aggregation.

        Args:
            pivot_cte_columns: The columns object from the pivot CTE

        Returns:
            Label: A labeled SQLAlchemy expression

        Raises:
            ValueError: If the field is not found in the pivot CTE
        """
        field_alias = self.field_to_alias(self.field)

        if not hasattr(pivot_cte_columns, field_alias):
            raise ValueError(f"Field '{self.field}' (alias: '{field_alias}') not found in pivot CTE columns")

        col = getattr(pivot_cte_columns, field_alias)

        numeric_col = cast(col, Integer)

        match self.type:
            case AggregationType.SUM:
                return func.sum(numeric_col).label(self.alias)
            case AggregationType.AVG:
                return func.avg(numeric_col).label(self.alias)
            case AggregationType.MIN:
                return func.min(numeric_col).label(self.alias)
            case AggregationType.MAX:
                return func.max(numeric_col).label(self.alias)
            case _:
                raise ValueError(f"Unsupported aggregation type: {self.type}")


Aggregation: TypeAlias = Annotated[CountAggregation | FieldAggregation, Field(discriminator="type")]



# Copyright 2019-2025 SURF, GÉANT.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Any

import structlog
from sqlalchemy import func, select
from sqlalchemy.orm import Query

from orchestrator.db import db
from orchestrator.domain.context_cache import cache_subscription_models
from orchestrator.search.core.types import EntityType
from orchestrator.search.indexing.indexer import Indexer
from orchestrator.search.indexing.registry import ENTITY_CONFIG_REGISTRY

logger = structlog.get_logger(__name__)


def _get_entity_count(stmt: Any) -> int | None:
    """Get total count of entities from a select statement."""

    count_stmt = select(func.count()).select_from(stmt.subquery())
    return db.session.execute(count_stmt).scalar()


def run_indexing_for_entity(
    entity_kind: EntityType,
    entity_id: str | None = None,
    dry_run: bool = False,
    force_index: bool = False,
    chunk_size: int = 1000,
    show_progress: bool = False,
) -> None:
    """Stream and index entities for the given kind.

    Builds a streaming query via the entity's registry config, disables ORM eager
    loads when applicable and delegates processing to `Indexer`.

    Args:
        entity_kind (EntityType): The entity type to index (must exist in
            `ENTITY_CONFIG_REGISTRY`).
        entity_id (Optional[str]): If provided, restricts indexing to a single
            entity (UUID string).
        dry_run (bool): When True, runs the full pipeline without performing
            writes or external embedding calls.
        force_index (bool): When True, re-indexes all fields regardless of
            existing hashes.
        chunk_size (int): Number of rows fetched per round-trip and passed to
            the indexer per batch.
        show_progress (bool): When True, logs progress for each processed entity.

    Returns:
        None
    """
    config = ENTITY_CONFIG_REGISTRY[entity_kind]

    q = config.get_all_query(entity_id)

    if isinstance(q, Query):
        q = q.enable_eagerloads(False)
        stmt = q.statement
    else:
        stmt = q

    total_count = _get_entity_count(stmt) if show_progress else None

    stmt = stmt.execution_options(stream_results=True, yield_per=chunk_size)
    entities = db.session.execute(stmt).scalars()

    indexer = Indexer(
        config=config,
        dry_run=dry_run,
        force_index=force_index,
        chunk_size=chunk_size,
        show_progress=show_progress,
        total_count=total_count,
    )

    with cache_subscription_models():
        indexer.run(entities)



# Copyright 2019-2025 SURF, GÉANT.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


class SearchUtilsError(Exception):
    """Base exception for this module."""

    pass


class ProductNotInRegistryError(SearchUtilsError):
    """Raised when a product is not found in the model registry."""

    pass


class ModelLoadError(SearchUtilsError):
    """Raised when a Pydantic model fails to load from a subscription."""

    pass


class InvalidCursorError(SearchUtilsError):
    """Raised when cursor cannot be decoded."""

    pass


class QueryStateNotFoundError(SearchUtilsError):
    """Raised when a query state cannot be found in the database."""

    pass



from uuid import UUID

from orchestrator.domain.base import SubscriptionModel
from orchestrator.services.subscriptions import _generate_etag, build_domain_model, build_extended_domain_model


async def get_subscription_dict(subscription_id: UUID, inject_inuseby: bool = True) -> tuple[dict, str]:
    """Helper function to get subscription dict by uuid from db or cache."""

    subscription_model = SubscriptionModel.from_subscription(subscription_id)

    if not inject_inuseby:
        subscription = build_domain_model(subscription_model)
    else:
        subscription = build_extended_domain_model(subscription_model)
    etag = _generate_etag(subscription)
    return subscription, etag



# Copyright 2019-2025 SURF.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from typing import Annotated, Any, Literal, Union
from uuid import UUID

from pydantic import BaseModel, Field, TypeAdapter

SCHEDULER_Q_CREATE = "create"
SCHEDULER_Q_UPDATE = "update"
SCHEDULER_Q_DELETE = "delete"


class APSchedulerJob(BaseModel):
    scheduled_type: Literal["create", "update", "delete"] = Field(..., description="Discriminator for job type")


class APSchedulerJobCreate(APSchedulerJob):
    name: str | None = Field(None, description="Human readable name e.g. 'My Process'")
    workflow_name: str = Field(..., description="Name of the workflow to run e.g. 'my_workflow_name'")
    workflow_id: UUID = Field(..., description="UUID of the workflow associated with this scheduled task")

    trigger: Literal["interval", "cron", "date"] = Field(..., description="APScheduler trigger type")
    trigger_kwargs: dict[str, Any] = Field(
        default_factory=lambda: {},
        description="Arguments passed to the trigger on job creation",
        examples=[{"hours": 12}, {"minutes": 30}, {"days": 1, "hours": 2}],
    )

    scheduled_type: Literal["create"] = Field("create", frozen=True)


class APSchedulerJobUpdate(APSchedulerJob):
    name: str | None = Field(None, description="Human readable name e.g. 'My Process'")
    schedule_id: UUID = Field(..., description="UUID of the scheduled task")

    trigger: Literal["interval", "cron", "date"] | None = Field(None, description="APScheduler trigger type")
    trigger_kwargs: dict[str, Any] | None = Field(
        default=None,
        description="Arguments passed to the job function",
        examples=[{"hours": 12}, {"minutes": 30}, {"days": 1, "hours": 2}],
    )

    scheduled_type: Literal["update"] = Field("update", frozen=True)


class APSchedulerJobDelete(APSchedulerJob):
    workflow_id: UUID = Field(..., description="UUID of the workflow associated with this scheduled task")
    schedule_id: UUID | None = Field(None, description="UUID of the scheduled task")

    scheduled_type: Literal["delete"] = Field("delete", frozen=True)


APSchedulerJobs = Annotated[
    Union[
        APSchedulerJobCreate,
        APSchedulerJobUpdate,
        APSchedulerJobDelete,
    ],
    Field(discriminator="scheduled_type"),
]
APSJobAdapter = TypeAdapter(APSchedulerJobs)  # type: ignore
